import numbers
import os
import pickle

from dateutil.parser import parse
from transformers import BertTokenizer, BertModel
import torch
from sentence_transformers import SentenceTransformer
from nltk.corpus import wordnet as wn

from pyhasura.synonyms import PartOfSpeech, is_part_of_speech

# Load a pre-trained model (e.g., 'all-MiniLM-L6-v2')
sentence_model = SentenceTransformer("all-MiniLM-L6-v2")
tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
model = BertModel.from_pretrained('bert-large-uncased')


def vectorize_string_mean(sentence):
    tokens = tokenizer.encode(sentence, add_special_tokens=True)
    input_ids = torch.tensor(tokens).unsqueeze(0)
    with torch.no_grad():
        outputs = model(input_ids)
        sentence_vector = outputs[0][:, 0, :].mean(dim=1)
        return sentence_vector.item()


def vectorize_string(sentence):
    return sentence_model.encode(sentence)


def vectorize_date_string(value):
    if not is_numeric(value):
        try:
            # Parse the RFC 3339 timestamp into a datetime object (including timezone)
            dt = parse(value)
            # Convert to an epoch timestamp (in seconds)
            return int(dt.timestamp())
        except ValueError:
            # If parsing fails, return the original value
            return value
    return value


def is_numeric(value):
    return isinstance(value, numbers.Number)


def is_boolean(value):
    return isinstance(value, bool)


def is_dictionary(value):
    return isinstance(value, dict)


def is_list(value):
    return isinstance(value, list)


def vectorize_dict(dictionary):
    for key, value in dictionary.items():
        if is_numeric(value):
            dictionary[key] = value
        elif is_dictionary(value):
            dictionary[key] = vectorize_dict(value)
        elif is_list(value):
            dictionary[key] = vectorize_dicts(value)
        elif is_boolean(value):
            if value:
                dictionary[key] = 1.0
            else:
                dictionary[key] = 0.0
        else:
            new_value = vectorize_date_string(value)
            if not is_numeric(new_value):
                new_value = vectorize_string_mean(value)
                dictionary[key] = new_value
    return dictionary


def vectorize_dicts(array_of_dicts):
    return list(map(lambda x: vectorize_dict(dict(x)), array_of_dicts))


def cosine_similarity(a, b):
    return torch.nn.functional.cosine_similarity(a, b).item()


if os.path.exists('words.pkl'):
    with open('words.pkl', 'rb') as f:
        words = pickle.load(f)
else:
    words = list(filter(lambda x: x.isalpha() and len(x) > 1, list(wn.all_lemma_names())))
    # Save the generated embeddings to 'embeddings.pkl'
    with open('words.pkl', 'wb') as f:
        pickle.dump(words, f)

if os.path.exists('inputs.pkl'):
    with open('inputs.pkl', 'rb') as f:
        input_ids_list = pickle.load(f)
else:
    input_ids_list = [tokenizer.encode(word, add_special_tokens=True, return_tensors="pt") for word in words]
    # Save the generated embeddings to 'embeddings.pkl'
    with open('inputs.pkl', 'wb') as f:
        pickle.dump(input_ids_list, f)

if os.path.exists('embeddings.pkl'):
    with open('embeddings.pkl', 'rb') as f:
        embeddings_list = pickle.load(f)
else:
    with torch.no_grad():
        embeddings_list = [model(input_ids).last_hidden_state.mean(dim=1) for input_ids in input_ids_list]

    # Save the generated embeddings to 'embeddings.pkl'
    with open('embeddings.pkl', 'wb') as f:
        pickle.dump(embeddings_list, f)


def find_related_words(target_word, part_of_speech=PartOfSpeech.Noun, top_n=5):
    target_word_embedding = model(
        tokenizer.encode(target_word, add_special_tokens=True, return_tensors="pt")).last_hidden_state.mean(dim=1)

    similarity_scores = [(word, cosine_similarity(target_word_embedding, emb)) for word, emb in
                         zip(words, embeddings_list)]

    # Sort by similarity
    similarity_scores = list(filter(lambda x: x[1] >= .8 and is_part_of_speech(x[0], part_of_speech), similarity_scores))
    sorted_similar_words = sorted(similarity_scores, key=lambda x: x[1], reverse=True)

    # Select top-N similar words
    similar_words = [word for word, _ in sorted_similar_words[:top_n]]

    return similar_words



from pluralizer import Pluralizer

pluralizer = Pluralizer()


def combine_words(word_lists, idx: int = 0):
    new_words = []
    if len(word_lists) == idx + 1:
        return list(set(
            [pluralizer.singular(word) for word in list(word_lists[idx])] +
            [pluralizer.plural(word) for word in word_lists[idx]]
        ))
    for other_word in combine_words(word_lists, idx + 1):
        for word in word_lists[idx]:
            new_words.append(pluralizer.plural(word) + ' ' + other_word)
            new_words.append(pluralizer.singular(word) + ' ' + other_word)
    return list(set(new_words))






from pydbml import PyDBML
from pathlib import Path
import camelcaser as cc
from pluralizer import Pluralizer

pluralizer = Pluralizer()


def add_dbml(dbml_file, kind, configuration, logging, hasura_client, customization):
    logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.DEBUG)
    parsed = PyDBML(Path(dbml_file))
    source = {
        "name": cc.make_camel_case(Path('global-retail-sales.dbml').stem),
        "kind": kind,
        "tables": [],
        "configuration": configuration,
        "customization": customization
    }

    rel_types = {
        "<": ("array", "object"),
        ">": ("object", "array"),
        "<>": ("array", "array"),
        "-": ("object", "object")
    }

    def create_table_metadata(dbml_table):
        return {
            "table": {
                "name": dbml_table.name,
                "schema": dbml_table.schema
            },
            "configuration": {
                "column_config": {},
                "custom_column_names": {},
                "custom_name": cc.make_camel_case(dbml_table.name),
                "custom_root_fields": {},
                "comment": dbml_table.note.text
            },
            "object_relationships": [],
            "array_relationships": []
        }

    def create_column_metadata(table_metadata, dbml_column):
        table_metadata["configuration"]["column_config"][dbml_column.name] = {
            "custom_name": cc.make_lower_camel_case(dbml_column.name),
            "comment": dbml_column.note.text
        }
        table_metadata["configuration"]["custom_column_names"][dbml_column.name] = cc.make_lower_camel_case(
            dbml_column.name)
        return table_metadata

    def create_relationship_metadata(dbml_ref, reverse=False):
        column_mapping = {}
        if reverse:
            for index, dbml_column in enumerate(dbml_ref.col2):
                column_mapping[dbml_column.name] = dbml_ref.col1[index].name
        else:
            for index, dbml_column in enumerate(dbml_ref.col1):
                column_mapping[dbml_column.name] = dbml_ref.col2[index].name
        return {
            "name": dbml_ref.name or cc.make_lower_camel_case(
                pluralizer.plural(dbml_ref.table1.name if reverse else dbml_ref.table2.name)),
            "comment": dbml_ref.comment,
            "using": {
                "manual_configuration": {
                    "column_mapping": column_mapping,
                    "insertion_order": None,
                    "remote_table": {
                        "name": dbml_ref.table1.name if reverse else dbml_ref.table2.name,
                        "schema": dbml_ref.table1.schema if reverse else dbml_ref.table2.schema,
                    }
                }
            }
        }

    for table in parsed.tables:
        new_table = create_table_metadata(table)
        for column in table.columns:
            create_column_metadata(new_table, column)
        source["tables"].append(new_table)
    for ref in parsed.refs:
        forward_rel_type = str(rel_types[ref.type][0])
        reverse_rel_type = str(rel_types[ref.type][1])
        forward_rel = create_relationship_metadata(ref)
        reverse_rel = create_relationship_metadata(ref, reverse=True)
        forward_table = [table for table in source["tables"] if table["table"]["name"] == ref.table1.name][0]
        forward_table[f"{forward_rel_type}_relationships"].append(forward_rel)
        reverse_table = [table for table in source["tables"] if table["table"]["name"] == ref.table2.name][0]
        reverse_table[f"{reverse_rel_type}_relationships"].append(reverse_rel)

    metadata = hasura_client.get_metadata()
    metadata["sources"].append(source)

    return metadata



import hashlib
import re
import copy
import camelcaser as cc
from neo4j import GraphDatabase
from pluralizer import Pluralizer
from graphql import is_object_type, is_non_null_type, is_list_type, is_leaf_type, is_enum_type, is_wrapping_type, \
    get_named_type
from sklearn_extra.cluster import KMedoids

from pyhasura.helpers import get_ordinal_of_smallest_number, compute_deltas
from pyhasura.keys import combine_words
from pyhasura.synonyms import split_cased_phrase, get_last_noun_sequence, get_synonyms
from pyhasura.vectorize import vectorize_string, find_related_words
import diskcache

# Specify the path where you want to store the cache
cache_path = ".cache/pyhasura/relationships"

# Create or load the cache
cache = diskcache.Cache(cache_path)

pluralizer = Pluralizer()

fk_candidates = {}
ak_candidates = {}
correlation_test = {}


class Relationships:
    def __init__(self, metadata, schema,
                 disallowed_key_types=None,
                 entity_synonyms=None,
                 alternate_keys=None,
                 key_suffixes=None,
                 logging_=None,
                 client=None,
                 key_uniqueness_test_size=10000,
                 key_uniqueness_test_rate=.2,
                 key_correlation_test_size=300000,
                 key_correlation_test_rate=.05
                 ):
        self.logging = logging_
        self.metadata = copy.deepcopy(metadata)
        self.schema = schema
        self.disallowed_key_types = disallowed_key_types or ['timestamptz', 'Boolean', 'timestamp']
        self.entity_synonyms = entity_synonyms or {}
        self.alternate_keys = alternate_keys or ['state', 'city', 'country', 'zip code', 'postal code']
        self.alternate_keys = self._get_cached_data(self.alternate_keys)
        self.alternate_keys_regex = [re.compile(ak, re.IGNORECASE) for ak in self.alternate_keys]
        self.key_suffixes = key_suffixes or ['id', 'key', 'serial number', 'code']
        self.relationships = []
        self.valid_keys = set()
        self.client = client
        self.key_uniqueness_test_size = key_uniqueness_test_size
        self.key_uniqueness_test_rate = key_uniqueness_test_rate
        self.key_correlation_test_size = key_correlation_test_size
        self.key_correlation_test_rate = key_correlation_test_rate
        self.type_names = [source.get('customization', {}).get('type_names') for source in
                           self.metadata.get('sources', [])
                           if source.get('customization', {}).get('type_names') is not None]

        def get_prefix_object_names(source):
            root_fields = source.get('customization', {}).get('root_fields', {})
            # prefix = root_fields.get('prefix', '')
            # suffix = root_fields.get('suffix', '')
            namespace = root_fields.get('namespace', '')
            return [
                f'{namespace}_mutation_frontend',
                f'{namespace}_query',
                f'{namespace}_subscription'
            ]

        self.root_fields = []
        for source in self.metadata.get('sources', []):
            if source.get('customization', {}).get('root_fields') is not None:
                self.root_fields += get_prefix_object_names(source)
        self.valid_keys_uniqueness = {}

    def _get_data_from_source(self, alternate_keys):

        self.logging.info('Generating the alternate key cache. This may take several minutes on first run.')
        candidates = [candidate.split() for candidate in alternate_keys]
        word_lists = []

        for candidate in candidates:
            word_list = []
            for word in candidate:
                synonyms = [synonyms.replace('_', ' ').lower() for synonyms in get_synonyms(word.lower())]
                synonyms.append(word.lower())
                synonyms = list(set(synonyms))
                word_list.append(synonyms)
            combined = combine_words(word_list)
            word_lists.append(combined)
        word_lists = sum(word_lists, [])
        result = ['|'.join(list({cc.make_snake_case(ak), cc.make_lower_camel_case(ak), ak})) for ak in word_lists]
        return result

    def _get_cached_data(self, alternate_keys):
        input_key = hashlib.sha256(bytes("\n".join(alternate_keys), "utf-8")).hexdigest()
        data = cache.get(input_key)
        # Check if data is already cached
        if data is None:
            # Derive data and cache it
            data = self._get_data_from_source(alternate_keys)
            result = cache.set(input_key, data)
            test = cache.get(input_key)
        return data

    def _create_pk_candidates(self, entity):
        schema = self.schema
        key_hints = self.key_suffixes
        phrase = split_cased_phrase(self.remove_type_names(entity))
        assumed_name = get_last_noun_sequence(phrase)
        similar_words = [{word.lower()} for word in assumed_name]
        for index, word_list in enumerate(similar_words):
            original = next(iter(word_list))
            similar_words[index] = word_list.union(
                set([word.replace('_', ' ') for word in get_synonyms(original)])).union(
                set([word.replace('_', ' ') for word in find_related_words(original)]))
        similar_words.append(key_hints)
        add_key_hints = list(set(
            [pluralizer.singular(hint) for hint in key_hints] +
            [pluralizer.plural(hint) for hint in key_hints]
        ))
        candidates = [candidate.split() for candidate in combine_words(similar_words)]
        if self.entity_synonyms.get(entity):
            candidates = candidates + \
                         [candidate.split() for candidate in
                          combine_words([self.entity_synonyms.get(entity), key_hints])]
        candidates_regexes = (
                ['^.*(' + ').*('.join(phrase) + ').*$' for phrase in candidates] +
                ['^(' + hint + ')$' for hint in add_key_hints])
        _regex = '|'.join(candidates_regexes)
        regex = re.compile(_regex, re.IGNORECASE)
        fields = schema.type_map[entity].fields.keys()
        possible_keys = set()
        for field in fields:
            if regex.match(field):
                possible_keys.add(field)
        return list(possible_keys)

    def _create_fk_candidates(self, primary_entity, foreign_entity, schema):
        key_hints = self.key_suffixes
        if fk_candidates.get(primary_entity) is None:
            phrase = split_cased_phrase(self.remove_type_names(primary_entity))
            assumed_name = get_last_noun_sequence(phrase)
            similar_words = [{word.lower()} for word in assumed_name]
            for index, word_list in enumerate(similar_words):
                original = next(iter(word_list))
                similar_words[index] = word_list.union(
                    set([word.replace('_', ' ') for word in get_synonyms(original)])).union(
                    set([word.replace('_', ' ') for word in find_related_words(original)]))
            similar_words.append(key_hints)
            candidates = [candidate.split() for candidate in combine_words(similar_words)]
            candidates += [[phrase] for phrase in key_hints]
            candidates_regexes = ['^.*(' + ').*('.join(phrase) + ').*$' for phrase in candidates]
            _regex = '|'.join(candidates_regexes)
            regex = re.compile(_regex, re.IGNORECASE)
            fk_candidates[primary_entity] = regex
        else:
            regex = fk_candidates[primary_entity]

        fields = [field[0] for field in schema.type_map[foreign_entity].fields.items() if is_leaf_type(field[1].type)]
        possible_keys = set()
        for field in fields:
            if regex.match(field):
                possible_keys.add(field)
        return list(possible_keys)

    def _create_ak_candidates(self, primary_entity, foreign_entity):
        possible_keys = set()
        if self.alternate_keys:
            schema = self.schema
            primary_fields = [field[0] for field in schema.type_map[primary_entity].fields.items() if
                              is_leaf_type(field[1].type)]
            foreign_fields = [field[0] for field in schema.type_map[foreign_entity].fields.items() if
                              is_leaf_type(field[1].type)]
            for regex in self.alternate_keys_regex:
                for foreign_field in foreign_fields:
                    for primary_field in primary_fields:
                        if regex.match(foreign_field) and regex.match(primary_field):
                            possible_keys.add((primary_entity, primary_field, foreign_entity, foreign_field))
        return list(possible_keys)

    def _get_leaf_type(self, t):
        if not is_leaf_type(t):
            return self._get_leaf_type(t.of_type)
        return t

    def _get_concrete_type(self, t):
        if is_wrapping_type(t):
            return self._get_concrete_type(t.of_type)
        return t

    def _is_list_type(self, t):
        if is_wrapping_type(t):
            if is_list_type(t):
                return True
            return self._is_list_type(t.of_type)
        return False

    def _is_relationship(self, t):
        if is_non_null_type(t):
            return self._is_relationship(t.of_type)
        return is_list_type(t) or is_object_type(t)

    @staticmethod
    def remove_prefix_and_suffix(text, prefix, suffix):
        # Construct a regular expression pattern to match the prefix and suffix
        pattern = f"^{re.escape(prefix)}(.*?){re.escape(suffix)}$"

        # Use re.sub() to replace the matched pattern with an empty string
        result = re.sub(pattern, r"\1", text)

        return result

    def remove_type_names(self, t):
        x = str(t)
        for type_names in self.type_names:
            x = self.remove_prefix_and_suffix(x, type_names.get('prefix', ''), type_names.get('suffix', ''))
        return x

    def _field_documents(self, objects):
        fields = []
        for key, value in objects:
            for field_key, field_value in value.fields.items():
                document = self.remove_type_names(key) + ' ' + str(value.description or '') + ' ' + field_key + ' ' + \
                           str(field_value.description or '')
                document = document.replace('_', ' ')
                vector = vectorize_string(document)
                if not self._is_relationship(field_value.type):
                    fields.append(
                        tuple([
                            key,
                            field_key,
                            document,
                            vector
                        ]))
        return fields

    def _check_object_type(self, item):
        root = self.remove_type_names(item[0])
        return is_object_type(item[1]) and not root.startswith('_') and not root.endswith('_aggregate') and not \
            root.endswith('_root') and not root.endswith('_fields') and not root.endswith('_response') and \
            root not in self.root_fields

    @staticmethod
    def _get_entity_name(source, table):
        return table.get('configuration', {}).get('custom_name') or \
            (source.get('customization', {}).get('type_names', {}).get('prefix', '') + table.get('table', {}).get(
                'name') + source.get('customization', {}).get('type_names', {}).get('suffix', ''))

    def _get_table(self, name=None, table_id=None):
        if name is not None:
            for source in self.metadata.get('sources', []):
                for table in source.get('tables', []):
                    test_name = self._get_entity_name(source, table)
                    if name == test_name:
                        return source, table
        elif table_id is not None:
            for source in self.metadata.get('sources', []):
                for table in source.get('tables', []):
                    if table.get('table', {}) == table_id:
                        return source, table
        return None, None

    def _metadata_tables(self):
        table_list = []
        for source in self.metadata.get('sources', []):
            for table in source.get('tables', []):
                name = self._get_entity_name(source, table)
                table_list.append(tuple([table.get('table'), name]))
        return table_list

    # Relationship: ((primary_table, primary_column, remote_table, remote_column),
    # name, existing|candidate, object|array)
    def create_existing_relationship(self, relationship_type, source, table):
        relationships = []
        for type_of_relationship in table.get(f'{relationship_type}_relationships', []):
            manual_configuration = type_of_relationship.get('using', {}).get('manual_configuration', {})
            remote_table_id = manual_configuration.get('remote_table')
            if remote_table_id is not None:
                column_mapping = list(manual_configuration.get('column_mapping', {}).items())
                remote_source, remote_table = self._get_table(table_id=remote_table_id)
                relationship = tuple([
                    tuple([self._get_entity_name(source, table),
                           column_mapping[0][0],
                           self._get_entity_name(source, remote_table),
                           column_mapping[0][1]]),
                    type_of_relationship.get('name', ''),
                    'existing',
                    relationship_type
                ])
                relationships.append(relationship)
            elif type_of_relationship.get('using', {}).get('foreign_key_constraint_on') is not None:
                self.logging.info('found foreign key constraint')
        if len(relationships) > 0:
            return relationships
        return None

    @staticmethod
    def _list_relationships(table):
        relationships = []
        for relationship_type in ['object', 'array', 'remote']:
            for type_of_relationship in table.get(f'{relationship_type}_relationships', []):
                relationships.append(type_of_relationship.get('name'))
        if len(relationships) > 0:
            return relationships
        return None

    def _get_key_guess_relationships(self):
        relationships = []
        rel_pks = {}
        for object_type in self._get_object_types():
            object_name = object_type[0]
            for remote_object_type in self._get_object_types():
                remote_object_name = remote_object_type[0]
                if remote_object_name != object_name and (object_name, remote_object_name) not in relationships:
                    relationships.append((object_name, remote_object_name))
        self.logging.info(f'Looking for PK/FK in {len(relationships)} combinations')
        for index, relationship in enumerate(relationships):
            pks = rel_pks.get(relationship[0]) if rel_pks.get(relationship[0]) is not None else (
                self._create_pk_candidates(relationship[0]))
            rel_pks[relationship[0]] = pks
            fks = self._create_fk_candidates(relationship[0], relationship[1], self.schema)
            if len(fks) == 0:
                self.logging.info(
                    f'{index + 1}-Entities {relationship[0]} and {relationship[1]} have no potential foreign keys')
            for pk in pks:
                for fk in fks:
                    new_rel = (relationship[0], pk, relationship[1], fk)
                    existing_relationships = [x[0] for x in self.relationships]
                    if new_rel not in existing_relationships:
                        rel = (new_rel, None, 'candidate', None)
                        self.logging.info(f'{index + 1}-Added relationship: {rel}')
                        self.relationships.append(rel)
            aks = self._create_ak_candidates(relationship[0], relationship[1])
            for rel in aks:
                self.logging.info(f'{index + 1}-Added alternate key relationship: {rel}')
                self.relationships.append((rel, None, 'ak', None))

    def _get_existing_relationships(self, tables):
        rels = []
        for table_id in tables:
            source, table = self._get_table(table_id[1])
            r = ((self.create_existing_relationship('object', source, table) or []) +
                 (self.create_existing_relationship('array', source, table) or []))
            rels = rels + list(filter(lambda x: x is not None, r))
        return rels

    def _get_object_types(self):
        return list(filter(self._check_object_type, self.schema.type_map.items()))

    def _get_semantic_relationships(self):
        relationships = []
        object_types = self._get_object_types()
        field_documents = self._field_documents(object_types)
        field_vectors = list(map(lambda x: x[3], field_documents))
        costs = []
        for k in range(2, min(50, len(field_vectors))):
            kmedoids = KMedoids(n_clusters=k, random_state=42)
            kmedoids.fit(field_vectors)
            costs.append(kmedoids.inertia_)
        optimum_clusters = get_ordinal_of_smallest_number(compute_deltas(costs)) + 2

        kmedoids = KMedoids(n_clusters=optimum_clusters, random_state=42)
        kmedoids.fit(field_vectors)
        cluster_labels = kmedoids.labels_
        medoids_indices = kmedoids.medoid_indices_
        for i, cluster_label in enumerate(cluster_labels):
            pk = field_documents[medoids_indices[cluster_label]]
            fk = field_documents[i]
            pk_type = self._get_leaf_type(self.schema.type_map[pk[0]].fields[pk[1]].type)
            fk_type = self._get_leaf_type(self.schema.type_map[fk[0]].fields[fk[1]].type)
            if pk[0] != fk[0] and pk_type.name == fk_type.name and pk_type.name not in self.disallowed_key_types \
                    and not is_enum_type(pk_type):
                forward_id = tuple([pk[0], pk[1], fk[0], fk[1]])
                backward_id = tuple([fk[0], fk[1], pk[0], pk[1]])
                already_exists_forward = len(list(filter(lambda x: x[0] == forward_id, relationships))) > 0
                already_exists_backward = len(list(filter(lambda x: x[0] == backward_id, relationships))) > 0
                if not already_exists_forward and not already_exists_backward:
                    relationship = (forward_id, None, 'candidate', None)
                    self.logging.info(f'Added relationship: {relationship}')
                    relationships.append(relationship)
        if len(relationships) > 0:
            self.relationships = self.relationships + relationships

    # Split by underscores (snake_case) and camel case boundaries
    @staticmethod
    def _extract_words(phrase):
        words = re.findall(r'[A-Z][a-z]*|[a-z]+', phrase)
        return words

    def _find_query_for_entity(self, entity):
        for query_name, query in self.schema.query_type.fields.items():
            if get_named_type(query.type).name == entity:
                return query_name
        return None

    def _get_timestamp_field(self, entity):
        return [name[0] for name in self.schema.type_map.get(entity).fields.items() if
                is_leaf_type(name[1].type) and name[1].type.name in ['timestamp', 'timestampz']]

    def _validate_keys(self):
        keys = set()
        self.valid_keys = set()
        for relationship in self.relationships:
            if relationship[2] == 'ak':
                self.valid_keys.add((relationship[0][0], relationship[0][1]))
                self.valid_keys.add((relationship[0][2], relationship[0][3]))
            elif relationship[3] is None:
                keys.add((relationship[0][0], relationship[0][1]))
                keys.add((relationship[0][2], relationship[0][3]))
        for key in keys:
            entity = key[0]
            column = key[1]
            query = self._find_query_for_entity(entity)
            if query is not None:
                result = self.client.execute(f"""
                    query test_{entity}_uniqueness {{
                        {query}(limit:{self.key_uniqueness_test_size}) {{
                            {column}
                        }}
                    }}""")
                result = list(map(lambda x: x.get(column), result[self.remove_type_names(entity)]))
                unique_result = set(result)
                if self.valid_keys_uniqueness.get(entity) is None:
                    self.valid_keys_uniqueness[entity] = {}
                unique_ratio = len(unique_result) / len(result) if len(result) > 0 else -1
                self.valid_keys_uniqueness[entity][column] = unique_ratio
                if unique_ratio > self.key_uniqueness_test_rate:
                    self.logging.info(f'Added valid key: {key}')
                    self.valid_keys.add(key)
                else:
                    timestamps = self._get_timestamp_field(entity)
                    if timestamps:
                        result = self.client.execute(f"""
                                            query test_{entity}_uniqueness {{
                                                {query}(limit:{self.key_uniqueness_test_size}) {{
                                                    {timestamps[0]}
                                                }}
                                            }}""")
                        result = list(map(lambda x: x.get(timestamps[0]), result[self.remove_type_names(entity)]))
                        unique_result = set(result)
                        unique_ratio = len(unique_result) / len(result) if len(result) > 0 else -1
                        if unique_ratio <= self.key_uniqueness_test_rate:
                            self.logging.info(f'Added valid key: {key}')
                            self.valid_keys.add(key)
                        else:
                            self.logging.info(f'Blacklisted key: {key}')
                    else:
                        self.logging.info(f'Blacklisted key: {key}')
        return

    def _validate_relationships(self):
        new_relationships = []
        for relationship in self.relationships:
            key, name, existing, relationship_type = relationship
            entity1, field1, entity2, field2 = key
            query1 = self._find_query_for_entity(entity1)
            query2 = self._find_query_for_entity(entity2)
            if query1 is not None and query2 is not None:
                if correlation_test.get(entity1 + field1) is None:
                    key1 = self.client.execute(f"""
                        query test_{entity1}_correlation {{
                            {query1}(limit:{self.key_correlation_test_size}) {{
                                {field1}
                            }}
                        }}""")
                    key1 = set(map(lambda x: x.get(field1), key1[self.remove_type_names(entity1)]))
                    correlation_test[entity1 + field1] = key1
                else:
                    key1 = correlation_test[entity1 + field1]

                if correlation_test.get(entity2 + field2) is None:
                    key2 = self.client.execute(f"""
                        query test_{entity2}_correlation {{
                            {query2}(limit:{self.key_correlation_test_size}) {{
                                {field2}
                            }}
                        }}""")
                    key2 = set(map(lambda x: x.get(field2), key2[self.remove_type_names(entity2)]))
                    correlation_test[entity2 + field2] = key2
                else:
                    key2 = correlation_test[entity2 + field2]

                overlap = len(key1 & key2)
                self.logging.info(f'Validated relationship: {relationship}')
                new_relationships.append((
                    key, name, existing, relationship_type,
                    (overlap / len(key1), overlap / len(key2),
                     min(overlap / len(key1), overlap / len(key2)
                         ))))
        self.relationships = new_relationships

    def _get_field_names(self, entity):
        return self.schema.type_map.get(entity).fields.keys()

    def _fix_relationship_name_collisions(self, relationship):
        key, name, existing, relationship_type, correlation, relationship_names = relationship
        forward_rel, forward_rel_type, backward_rel, backward_rel_type = relationship_names
        entity1, field1, entity2, field2 = key
        source, table = self._get_table(name=entity1)
        existing_relationships = [name.lower() for name in self._list_relationships(table) or []]
        single = pluralizer.singular(forward_rel).lower()
        plural = pluralizer.plural(forward_rel).lower()
        if existing == 'existing' or single in existing_relationships or plural in existing_relationships:
            forward_rel = None
        else:
            field_names = [name.lower() for name in self._get_field_names(entity1)]
            if existing == 'ak' or single in field_names or pluralizer.plural(
                    forward_rel).lower() in field_names:
                rel = f'{forward_rel}By{field1}'
                if rel.lower() in field_names:
                    forward_rel = None
                else:
                    forward_rel = rel
        source, table = self._get_table(name=entity2)
        existing_relationships = [name.lower() for name in self._list_relationships(table) or []]
        single = pluralizer.singular(backward_rel).lower()
        plural = pluralizer.plural(backward_rel).lower()
        if single in existing_relationships or plural in existing_relationships:
            backward_rel = None
        else:
            field_names = [name.lower() for name in self._get_field_names(entity2)]
            if existing == 'ak' or single in field_names or pluralizer.plural(
                    backward_rel).lower() in field_names:
                rel = f'{backward_rel}By{field2}'
                if rel.lower() in field_names:
                    backward_rel = None
                else:
                    backward_rel = rel
        return forward_rel, forward_rel_type, backward_rel, backward_rel_type

    def _create_relationship_names(self):
        new_relationships = []
        for relationship in self.relationships:
            key, name, existing, relationship_type, correlation = relationship
            entity1, field1, entity2, field2 = key
            forward, reverse, maximum = correlation
            key1_unique = self.valid_keys_uniqueness.get(entity1).get(field1)
            key2_unique = self.valid_keys_uniqueness.get(entity2).get(field2)
            if maximum > self.key_correlation_test_rate:
                if key1_unique == 1 and key2_unique == 1:
                    relationship_names = (
                        pluralizer.singular(self.remove_type_names(entity2)), 'object',
                        pluralizer.singular(self.remove_type_names(entity1)), 'object')
                elif key2_unique == 1 and key1_unique != 1:
                    relationship_names = (pluralizer.singular(self.remove_type_names(entity2)), 'object',
                                          pluralizer.plural(self.remove_type_names(entity1)), 'array')
                elif key1_unique == 1 and key2_unique != 1:
                    relationship_names = (pluralizer.plural(self.remove_type_names(entity2)), 'array',
                                          pluralizer.singular(self.remove_type_names(entity1)), 'object')
                else:
                    relationship_names = (pluralizer.plural(self.remove_type_names(entity2)), 'array',
                                          pluralizer.plural(self.remove_type_names(entity1)), 'array')
                relationship_names = (
                    cc.make_lower_camel_case(relationship_names[0]), relationship_names[1],
                    cc.make_lower_camel_case(relationship_names[2]), relationship_names[3])
                relationship_names = self._fix_relationship_name_collisions(
                    (key, name, existing, relationship_type, correlation, relationship_names))
                new_relationships.append((key, name, existing, relationship_type, correlation, relationship_names))
        self.relationships = new_relationships

    def _get_column_name(self, entity, name):
        source, table = self._get_table(entity)
        fields = table.get('configuration', {}).get('column_config', {}).items()
        if len(fields) == 0:
            return name
        result = list(map(lambda x: x[0], filter(lambda x: x[1].get('custom_name') == name, fields)))
        if len(result) == 0:
            return name
        return result[0]

    def _update_metadata(self):
        for relationship in self.relationships:
            key, _, _, _, _, relationship_names = relationship
            entity1, field1, entity2, field2 = key
            forward_name, forward_type, reverse_name, reverse_type = relationship_names
            column1 = self._get_column_name(entity1, field1)
            column2 = self._get_column_name(entity2, field2)

            def add_relationship(_entity1, _entity2, _column1, _column2, relation_type, relation_name):
                source, table = self._get_table(_entity1)
                remote_source, remote_table = self._get_table(_entity2)
                if source == remote_source:
                    rels = table.get(f'{relation_type}_relationships') or []
                    if len([found for found in rels if found.get('name') == relation_name]) == 0:
                        if len(rels) == 0:
                            rels = []
                            table[f'{relation_type}_relationships'] = rels
                        rel = {
                            "name": relation_name,
                            "using": {
                                "manual_configuration": {
                                    "column_mapping": {
                                        _column1: _column2
                                    },
                                    "insertion_order": None,
                                    "remote_table": remote_table.get('table')
                                }
                            }
                        }
                        rels.append(rel)
                else:
                    rels = table.get('remote_relationships')
                    if rels is None:
                        rels = []
                        table['remote_relationships'] = rels
                    rel = {
                        "definition": {
                            "to_source": {
                                "field_mapping": {
                                    _column1: _column2
                                },
                                "relationship_type": relation_type,
                                "source": remote_source,
                                "table": remote_table.get('table')
                            }
                        },
                        "name": relation_name
                    }
                    rels.append(rel)

            if forward_name is not None:
                add_relationship(entity1, entity2, column1, column2, forward_type, forward_name)
            if reverse_name is not None:
                add_relationship(entity2, entity1, column2, column1, reverse_type, reverse_name)

    def get_schema_relationships(self):
        """
        Get the relationships between object types in the schema.

        Returns:
            nodes (set): A set of tuples containing the object types and their labels.
            relationships (list): A list of tuples containing the relationship information between object types.
        """
        relationships = []
        nodes = set()
        object_types = self._get_object_types()
        for object_type in object_types:
            object_type_source, object_type_table = self._get_table(name=object_type[0])
            object_type_labels = (object_type[0], object_type_source.get('name'), object_type_table.get('table').get('name'))
            nodes.add((object_type, object_type_labels))
            fields = object_type[1].fields.items()
            for field in fields:
                concrete_type = self._get_concrete_type(field[1].type)
                if is_object_type(concrete_type) and self._check_object_type((concrete_type.name, concrete_type)):
                    concrete_type_source, concrete_type_table = self._get_table(name=concrete_type.name)
                    concrete_type_labels = (
                        concrete_type.name, concrete_type_source.get('name'), concrete_type_table.get('table').get('name'))
                    nodes.add(((concrete_type.name, concrete_type), concrete_type_labels))
                    rel_type = 'array' if self._is_list_type(field[1].type) else 'object'
                    field_parts = field[0].split('By')
                    rel_modifier = field_parts[1] if len(field_parts) > 1 else None
                    rel_name = f'HAS_{pluralizer.plural(cc.make_snake_case(concrete_type.name)).upper()}' \
                        if rel_type == 'array' \
                        else f'HAVE_{pluralizer.singular(cc.make_snake_case(concrete_type.name)).upper()}'
                    if rel_modifier is not None:
                        rel_name += '_BY_' + rel_modifier.upper()
                    relationships.append((
                        object_type,
                        object_type_source,
                        object_type_table,
                        object_type_labels,
                        field,
                        rel_type,
                        rel_name,
                        (concrete_type.name, concrete_type),
                        concrete_type_source,
                        concrete_type_table,
                        concrete_type_labels
                    ))
        return nodes, relationships

    def analyze_relationships(self):
        self.logging.info('Analyzing relationships...')
        tables = self._metadata_tables()
        self.logging.info(f'Found {len(tables)} tables.')
        self.relationships = self._get_existing_relationships(tables) or []
        self.logging.info(f'Found {len(self.relationships)} relationships, existing.')
        self._get_semantic_relationships()
        self.logging.info(f'Found {len(self.relationships)} relationships, existing + semantic.')
        self._get_key_guess_relationships()
        self.logging.info(f'Found {len(self.relationships)} relationships, existing + semantic + key-guess.')
        # validate key uniqueness
        self._validate_keys()
        self.relationships = [relationship for relationship in self.relationships if
                              (relationship[0][0], relationship[0][1]) in self.valid_keys and
                              (relationship[0][2], relationship[0][3]) in self.valid_keys
                              ]
        self.logging.info(f'Found {len(self.relationships)} relationships, existing + semantic + key-guess - bad-keys.')
        # validate relationship correlation
        self._validate_relationships()
        self.logging.info(f'Validated {len(self.relationships)} relationships.')
        # generate relation names
        self._create_relationship_names()
        self.logging.info(f'Finalized {len(self.relationships)} relationships.')
        # add to hasura metadata
        self._update_metadata()
        # write out result
        return self.metadata

    def metadata_to_neo4j(self, uri, username, password, nodes=None, relationships=None):
        """
        Args:
            uri: The URI of the Neo4j database.
            username: The username used for authentication.
            password: The password used for authentication.
            nodes: A list of nodes to be created in the Neo4j database. Each node should be a tuple containing three elements: (fields, node_info, table_info). `fields` is a dictionary containing
        * the field names and their corresponding types. `node_info` is a tuple containing the name, source, and table information of the node. `table_info` is a dictionary containing additional
        * information about the table.
            relationships: A list of relationships to be created in the Neo4j database. Each relationship should be a tuple containing eleven elements: (_, _, _, source_labels, field, _, rel
        *_name, _, _, _, destination_labels). `_` represents unused elements. `source_labels` and `destination_labels` are tuples containing the name, source, and additional information of the
        * source and destination nodes, respectively. `field` is a tuple containing the name and additional information of the field. `rel_name` is the name of the relationship.

        Example usage:

        nodes = [
            ({"field1": "type1", "field2": "type2"}, ("Node1", "Source1", "Table1"), {}),
            ({"field3": "type3", "field4": "type4"}, ("Node2", "Source2", "Table2"), {}),
        ]

        relationships = [
            (_, _, _, ("Source1", "Node1", {}), ("Field1", {}), _, "RELATIONSHIP1", _, _, _, ("Source2", "Node2", {})),
            (_, _, _, ("Source2", "Node2", {}), ("Field2", {}), _, "RELATIONSHIP2", _, _, _, ("Source1", "Node1", {})),
        ]

        metadata_to_neo4j(uri="neo4j://localhost:7687", username="admin", password="password", nodes=nodes, relationships=relationships)
        """
        driver = GraphDatabase.driver(uri, auth=(username, password))

        def create_node(tx, node):
            name, source, table = node[1]
            fields = {"_name": name, "_source": source, "_table": table}

            for field in node[0][1].fields.items():
                concrete_type = self._get_concrete_type(field[1].type)
                field_name = field[0] if field[0] != 'name' else 'name_'
                if not is_object_type(concrete_type):
                    fields[field_name] = concrete_type.name
            tx.run(f"CREATE (a:{name}:{source}:DataProduct) SET a = $fields RETURN a",
                   fields=fields)

        def create_relationship(tx, relationship):
            _, _, _, source_labels, field, _, rel_name, _, _, _, destination_labels = relationship
            field_name, _ = field
            source_name, source_source, _ = source_labels
            dest_name, dest_source, _ = destination_labels
            tx.run(f"""
                MATCH (a:{source_name}:{source_source}:DataProduct), (b:{dest_name}:{dest_source}:DataProduct)
                CREATE (a)-[:{rel_name} {{relationshipName: $field_name}}]->(b)
            """, field_name=field_name)

        def remove_nodes(tx):
            tx.run(f"MATCH (n:DataProduct) DETACH DELETE n")

        with driver.session() as session:
            session.write_transaction(remove_nodes)
            for node in nodes:
                session.write_transaction(create_node, node)
            for relationship in relationships:
                session.write_transaction(create_relationship, relationship)

        driver.close()



import base64
import copy
import json
import os
import pickle
import tempfile
from enum import Enum

import pandas as pd
import pyarrow as pa
import requests
from gql import gql
from graphql import build_client_schema
from neo4j import GraphDatabase
from pymongo import MongoClient
from sklearn.ensemble import IsolationForest
from sklearn.feature_extraction import DictVectorizer
from sklearn_extra.cluster import KMedoids

from pyhasura import gql_client, flatten_nested_dicts
from pyhasura.helpers import get_ordinal_of_smallest_number, compute_deltas
from pyhasura.relationships import Relationships
from pyhasura.upload_csvs import import_tables, Casing
from pyhasura.vectorize import vectorize_dicts
from pyhasura.dbml_hasura import add_dbml
from gql.transport.requests import log as requests_logger


class ExportFormat(Enum):
    FLAT = 1
    DATAFRAME = 2
    ARROW = 3
    NATURAL = 4
    PARQUET = 5
    CSV = 6


def add_score(item, index, score):
    copy = dict(item)
    copy['__score__'] = score
    copy['__index__'] = index
    return copy


def add_dummy_features(dicts, n, m):
    for d in dicts:
        z = m
        while len(d) < n and z != 0:
            # Add dummy features with default values (e.g., 0)
            dummy_key = f'dummy_feature_{len(d)}'
            d[dummy_key] = 0
            z = z - 1
    return dicts


def create_empty_arrays(n):
    return [[] for _ in range(n)]


class HasuraClient:
    """

    """

    def __init__(self, uri=None, role=None, admin_secret=None, output_dir='.', headers=None, logging_=None):
        """
        Args:
            uri (str): The URI for the GraphQL server.
            role (str, optional): The role to be used for Hasura authorization. Defaults to None.
            admin_secret (str, optional): The Hasura admin secret. Defaults to None.
            output_dir (str, optional): The directory where output files will be saved. Defaults to '.'.
            headers (dict, optional): Additional headers to be included in the GraphQL request. Defaults to None.
            logging_ (bool, optional): Whether to enable logging. Defaults to None.

        """
        if headers is None:
            headers = {}
        if role is not None:
            headers['x-hasura-role'] = role
        if admin_secret is not None:
            headers['x-hasura-admin-secret'] = admin_secret
        self.uri = uri
        self.admin_secret = admin_secret
        if uri is not None:
            self.client = gql_client(self.uri, headers=headers)
        if logging_ is not None:
            self.logging = logging_
        self.result = {}
        self.native_result = {}
        self.vectorized_result = {}
        self.anomalies_result = {}
        self.output_dir = output_dir
        self.result_format = None
        self.operation_name = None
        self.extensions: dict = {}
        self.metadata = {}
        requests_logger.setLevel(self.logging.ERROR)

    def _transform_to_flatten(self):
        flatten_result = {key: flatten_nested_dicts(val) for key, val in self.native_result.items()}
        return flatten_result

    def _transform_to_dataframe(self):
        flatten_result = {key: pd.DataFrame(flatten_nested_dicts(val)) for key, val in self.native_result.items()}
        return flatten_result

    def _transform_to_arrow(self):
        flatten_result = {key: pa.Table.from_pandas(pd.DataFrame(flatten_nested_dicts(val))) for key, val in
                          self.native_result.items()}
        return flatten_result

    def execute(self, query, variables=None, output_format=ExportFormat.NATURAL, operation_name=None):
        op = gql(query)
        operation_name = operation_name if operation_name is not None else op.definitions[0].name.value
        result = self.client.execute(gql(query), variables, operation_name=operation_name, get_execution_result=True)
        self.extensions = result.extensions
        self.native_result = result.data
        self.result = self.native_result
        self.operation_name = operation_name
        return self.convert_output_format(output_format)

    def _get_file_path(self, extension):
        return os.path.join(self.output_dir, extension)

    @staticmethod
    def _write_result_as_file(file_name, result_data):
        with open(file_name, 'w') as file:
            json.dump(result_data, file, indent=4)

    def set_data(self, data, output_format=ExportFormat.NATURAL, operation_name=None):
        """
        Set the data for the current instance.

        Args:
            data: The data to be set for the instance.
            output_format: The output format to be used for the data (default is ExportFormat.NATURAL).
            operation_name: An identifier for the operation that generated the dataset (default is None)

        Returns:
            The converted data based on the specified output format.
        """
        self.native_result = data
        self.result = self.native_result
        self.operation_name = operation_name
        return self.convert_output_format(output_format)

    def get_metadata(self):
        """Gets the metadata for the given object.

        This method makes a request to the specified URI to retrieve the metadata for the object.
        It replaces the 'graphql' path with 'metadata' in the URI and sends a POST request with the
        admin secret header to get the metadata. The metadata is stored in the `metadata` attribute
        of the object and a deep copy of the metadata is returned.

        Returns:
            A deep copy of the metadata for the object.
        """
        self.metadata = json.loads(requests.post(self.uri.replace('graphql', 'metadata'),
                                                 headers={'x-hasura-admin-secret': self.admin_secret}).text)
        return copy.deepcopy(self.metadata)

    def reload_schema(self):
        """
        Reloads the schema by sending a GET request to the hasura proxy server.

        Returns:
            response (requests.Response): The response object containing the server's response.
        """
        return requests.get(self.uri.replace('graphql', 'reload-schema'))

    def relationship_analysis(
            self,
            filename=None,
            entity_synonyms=None,
            alternate_keys=None,
            key_suffixes=None,
            key_uniqueness_test_size=10000,
            key_uniqueness_test_rate=.2,
            key_correlation_test_size=300000,
            key_correlation_test_rate=.05,
    ):
        """
        Args:
            filename: The name of the file to save the metadata analysis results. If not specified, the results will not be saved to a file.
            entity_synonyms: A dictionary containing entity synonyms to be used in the analysis. Each key-value pair represents a synonym mapping, where the key is the original entity name and
        * the value is the synonym.
            alternate_keys: A list of alternate keys to consider in the analysis. Each alternate key is a list of entity attributes that should uniquely identify an entity.
            key_suffixes: A list of suffixes to add to the names of the keys to be tested for uniqueness and correlation. Each suffix should be a string.
            key_uniqueness_test_size: The number of entity records to sample for testing key uniqueness. Default is 10000.
            key_uniqueness_test_rate: The ratio at which to sample entity records for testing key uniqueness. Default is 0.2 (20%).
            key_correlation_test_size: The number of entity records to sample for testing key correlation. Default is 300000.
            key_correlation_test_rate: The ratio at which to sample entity records for testing key correlation. Default is 0.05 (5%).

        Returns:
            metadata: A dictionary containing the relationship analysis results.

        """
        rel = Relationships(
            self.get_metadata(),
            self.get_schema(),
            logging_=self.logging,
            client=self,
            alternate_keys=alternate_keys,
            key_suffixes=key_suffixes,
            entity_synonyms=entity_synonyms,
            key_uniqueness_test_size=key_uniqueness_test_size,
            key_uniqueness_test_rate=key_uniqueness_test_rate,
            key_correlation_test_size=key_correlation_test_size,
            key_correlation_test_rate=key_correlation_test_rate
        )
        metadata = rel.analyze_relationships()
        if filename is not None:
            with open(filename, 'w') as f:
                json.dump({"metadata": metadata}, f, indent=4)
        return metadata

    def get_schema(self):
        """
        Returns the schema of the client.

        :return: The schema of the client.
        :rtype: build_client_schema.graphql.GraphQLSchema
        """
        return build_client_schema(self.get_sdl())

    def get_sdl(self):
        """
        Retrieves the Schema Definition Language (SDL) for the GraphQL API.

        Returns:
            str: The Schema Definition Language (SDL) for the GraphQL API.

        Raises:
            Exception: If there is an error executing the introspection query.

        Example:
            >>> sdl = obj.get_sdl()
        """
        introspection_query = """
            query IntrospectionQuery {
                __schema {
                    queryType {
                        name
                    }
                    mutationType {
                        name
                    }
                    subscriptionType {
                        name
                    }
                    types {
                        ...FullType
                    }
                    directives {
                        name
                        description
                        locations
                        args {
                            ...InputValue
                        }
                    }
                }
            }
    
            fragment FullType on __Type {
                kind
                name
                description
                fields(includeDeprecated: true) {
                    name
                    description
                    args {
                        ...InputValue
                    }
                    type {
                        ...TypeRef
                    }
                    isDeprecated
                    deprecationReason
                }
                inputFields {
                    ...InputValue
                }
                interfaces {
                    ...TypeRef
                }
                enumValues(includeDeprecated: true) {
                    name
                    description
                    isDeprecated
                    deprecationReason
                }
                possibleTypes {
                    ...TypeRef
                }
            }
    
            fragment InputValue on __InputValue {
                name
                description
                type {
                    ...TypeRef
                }
                defaultValue
            }
    
            fragment TypeRef on __Type {
                kind
                name
                ofType {
                    kind
                    name
                    ofType {
                        kind
                        name
                        ofType {
                            kind
                            name
                            # Continue nesting for more complex types if needed
                        }
                    }
                }
            }
            """
        return self.execute(introspection_query)

    def vectorize_result(self, min_features_dict=None):
        """
        Vectorizes the last query result using DictVectorizer.

        :return: self.vectorized_result, a dictionary with the same keys as the last query, but with the values
                 transformed into dense arrays using DictVectorizer.
        """
        self.vectorized_result = {}
        self.logging.debug(f'Vectorizing: {json.dumps(self.native_result, indent=2)}')
        for key in self.native_result:
            v = DictVectorizer(sparse=False)  # Set sparse=False for a dense array
            flatted_dict = flatten_nested_dicts(self.native_result[key])
            vectorized = vectorize_dicts(flatted_dict)
            self.vectorized_result[key] = v.fit_transform(vectorized)
            rows, num_features = self.vectorized_result[key].shape
            if min_features_dict is not None and min_features_dict.get(key) is not None and \
                    num_features < min_features_dict.get(key):
                vectorized = add_dummy_features(vectorized, min_features_dict.get(key),
                                                min_features_dict.get(key) - num_features)
                self.vectorized_result[key] = v.fit_transform(vectorized)
            self.logging.info(self.vectorized_result[key])
        return self.vectorized_result

    def clusters(self, n_clusters):
        """
        :param n_clusters: A dictionary containing the number of clusters for each key
        :return: A dictionary containing the clustered data for each key

        Performs clustering on the vectorized result using the KMedoids algorithm.
        Clusters the data for each key separately and returns a dictionary containing the clustered data for each key.

        Example usage:
            self.clusters({
                'key1': 3,
                'key2': 5,
                ...
            })
        """
        self.vectorize_result()
        clustered = {}
        for key in self.vectorized_result:
            kmedoids = KMedoids(n_clusters=n_clusters[key], random_state=42)
            kmedoids.fit(self.vectorized_result[key])
            cluster_labels = kmedoids.labels_
            clustered[key] = create_empty_arrays(cluster_labels.max() + 1)
            # Print the cluster assignments
            for i, cluster_label in enumerate(cluster_labels):
                clustered[key][cluster_label].append(self.native_result[key][i])
        return clustered

    def optimal_number_of_clusters(self, lower, upper):
        """
        Find the optimal number of clusters for each key in the vectorized result.

        :param lower: The lower bound of the range of cluster numbers to consider.
        :param upper: The upper bound of the range of cluster numbers to consider.
        :return: A dictionary containing the optimal number of clusters for each key in the vectorized result.
        """
        self.vectorize_result()
        clustered = {}
        for key in self.vectorized_result:
            costs = []
            for k in range(lower, upper + 1):
                kmedoids = KMedoids(n_clusters=min(k, len(self.vectorized_result[key]) - 1), random_state=42)
                kmedoids.fit(self.vectorized_result[key])
                costs.append(kmedoids.inertia_)
            clustered[key] = get_ordinal_of_smallest_number(compute_deltas(costs)) + 2
        return clustered

    def anomalies_training(self, output_dir=None, base64_encoded_data=False, database_output=False,
                           selection_set_hash=None):
        """
        Trains anomaly detection models and saves them.

        Args:
            output_dir (str, optional): The directory where the models will be saved. If not provided, the default directory will be used.
            base64_encoded_data (bool, optional): Indicates whether the models should be stored as base64-encoded data. Default is False.
            database_output (bool, optional): Indicates whether the models should be stored in a database. Default is False.
            selection_set_hash (str, optional): The hash value for this selection set. Default is an empty string.

        Returns:
            dict: A dictionary containing filenames/path of the saved models. If database_output is True and a MONGODB_CONNECTION_STRING environment variable is set, saves the model.

        Notes:
            - Requires scikit-learn library (version 0.22.2 or higher).
            - If base64_encoded_data ais True, the models will be returned as base64-encoded data in the dictionary.
            - The output_dir will be created if it doesn't exist.
            - The saved models are in pickle (.pkl) format.
            - If database_output is True, the models will be saved in a MongoDB collection named 'anomalyDetectionModels'.
            - The connection string for the MongoDB database should be provided as the MONGODB_CONNECTION_STRING environment variable.

        """
        if output_dir is not None:
            self.output_dir = output_dir
        self.vectorize_result()
        clf = IsolationForest(contamination=0.1, random_state=42)
        filenames = {"selectionSetHash": selection_set_hash, "operationName": self.operation_name}
        for key in self.vectorized_result:
            clf.fit(self.vectorized_result[key])
            temp_dir = tempfile.mkdtemp()
            filename = os.path.join(self.output_dir, temp_dir, key + '.pkl')
            if base64_encoded_data or database_output:
                filenames[key] = base64.b64encode(pickle.dumps(clf)).decode('utf-8')
            else:
                with open(filename, 'wb') as model_file:
                    pickle.dump(clf, model_file)
                    filenames[key] = filename
        if database_output and os.environ.get("MONGODB_CONNECTION_STRING") is not None:
            connection_string = os.environ.get("MONGODB_CONNECTION_STRING")
            default_db_name = connection_string.split('/')[-1]
            client = MongoClient(connection_string)
            db = client[default_db_name]
            collection = db['anomalyDetectionModels']
            if selection_set_hash is not None:
                collection.update_one({"selectionSetHash": selection_set_hash}, {'$set': filenames}, upsert=True)
                client.close()
                return {"selectionSetHash": selection_set_hash}
            else:
                collection.update_one({"operationName": self.operation_name}, {'$set': filenames}, upsert=True)
                client.close()
                return {"operationName": self.operation_name}

        return filenames

    def anomalies(self, training_files=None, threshold=0, training_base64=None, selection_set_hash=None,
                  operation_name=None):
        """
        Args:
            training_files: A dictionary mapping keys to file paths for training files (default: None)
            threshold: The threshold value below which anomalies are considered (default: 0)
            training_base64: Base64 encoded training data (default: None)
            selection_set_hash: Hash value for selecting a specific set of training data (default: None)
            operation_name: Name of the operation for selecting training data (default: None)

        Returns:
            A dictionary containing anomalies found in the input data.
        """

        self.anomalies_result = {}
        # Initialize Isolation Forest
        trained_models = {}
        min_features = {}
        if selection_set_hash is not None:
            connection_string = os.environ.get("MONGODB_CONNECTION_STRING")
            default_db_name = connection_string.split('/')[-1]
            client = MongoClient(connection_string)
            db = client[default_db_name]
            collection = db['anomalyDetectionModels']
            training_base64 = collection.find_one({"selectionSetHash": selection_set_hash})
            if training_base64 is not None and training_base64.get('selectionSetHash') is not None:
                self.logging.info('Loaded trained model from selectionSetHash')
            else:
                self.logging.info('Problem loading trained model from selectionSetHash. Will generate model from '
                                  'input dataset.')
            client.close()
        elif operation_name is not None:
            connection_string = os.environ.get("MONGODB_CONNECTION_STRING")
            default_db_name = connection_string.split('/')[-1]
            client = MongoClient(connection_string)
            db = client[default_db_name]
            collection = db['anomalyDetectionModels']
            training_base64 = collection.find_one({"operationName": operation_name})
            if training_base64 is not None and training_base64.get('operationName') is not None:
                self.logging.info('Loaded trained model from operationName')
            else:
                self.logging.info('Problem loading trained model from operationName. Will generate model from '
                                  'input dataset.')
            client.close()
        for key in self.vectorized_result:
            if training_files is not None:
                with open(training_files[key], 'rb') as model_file:
                    trained_models[key] = pickle.load(model_file)
                    min_features[key] = trained_models[key].n_features_in_
            elif training_base64 is not None:
                trained_models[key] = pickle.loads(base64.b64decode(training_base64[key]))
                min_features[key] = trained_models[key].n_features_in_
        self.vectorize_result(min_features)
        for key in self.vectorized_result:
            if trained_models.get(key) is None:
                trained_models[key] = IsolationForest(contamination=0.1, random_state=42)
                trained_models[key].fit(self.vectorized_result[key])
            scores = trained_models[key].decision_function(self.vectorized_result[key])
            merged = list(map(lambda x: add_score(self.native_result[key][x[0]], x[0], x[1]), enumerate(scores)))
            self.anomalies_result[key] = list(filter(lambda x: x['__score__'] < threshold, merged))
        return self.anomalies_result

    def convert_output_format(self, output_format):
        """
        Convert the output format of the result.

        :param output_format: The desired output format. Must be one of ExportFormat.FLAT, ExportFormat.DATAFRAME, ExportFormat.PARQUET, ExportFormat.CSV, ExportFormat.ARROW.
        :return: The converted result in the specified format.
        """
        self.result_format = output_format
        if output_format == ExportFormat.FLAT:
            self.result = self._transform_to_flatten()
        elif output_format in [ExportFormat.DATAFRAME, ExportFormat.PARQUET, ExportFormat.CSV]:
            self.result = self._transform_to_dataframe()
        elif output_format == ExportFormat.ARROW:
            self.result = self._transform_to_arrow()
        return self.result

    def get_extensions(self):
        """
        Returns the dict of extensions.

        Args:
            self: The object itself.

        Returns:
            A dictionary of extensions.
        """
        return self.extensions

    def get_data(self):
        """
        Method Name:
        get_data

        Description:
        This method retrieves the native result from the object.

        Parameters:
        self - The current object instance.

        Returns:
        The native result of the object.
        """
        return self.native_result

    def get_data_frame(self):
        """
        Method: get_data_frame

        Description:
        This method converts the native_result attribute to a Pandas DataFrame object and returns it.

        Parameters:
        - self: The instance of the class calling the method.

        Return Type:
        - Pandas DataFrame: A DataFrame object containing the data from the native_result attribute.

        Example Usage:
        df = instance.get_data_frame()

        """
        return pd.DataFrame(self.native_result)

    def upload_csv_folder(self, csv_folder, uri, if_exists='replace', casing=Casing.none):
        """
        Uploads a folder of CSV files to a specified URI.

        Args:
            csv_folder (str): The path to the folder containing the CSV files.
            uri (str): The destination database URI where the CSV files will be uploaded.
            if_exists (str, optional): The action to take if a table with the same name already exists at the destination URI. Default is 'replace'.
            casing (Casing, optional): The casing rule to apply to the column names. Defaults to Casing.none.

        Returns:
            The table names that were imported from the CSV files.
        """
        return import_tables(csv_folder=csv_folder, uri=uri, if_exists=if_exists, casing=casing, logging=self.logging)

    def write_to_file(self, output_dir=None, output_format=None):
        """
        Writes the result data to file(s) in the specified output format and directory.

        :param output_dir: The directory where the output file(s) will be written. If not provided, the current output directory is used.
        :param output_format: The format in which the output file(s) will be written. If not provided, the current output format is used.
        :return: A list of filenames corresponding to the output file(s) created.

        """
        if output_dir is not None:
            self.output_dir = output_dir
        if output_format is not None:
            self.convert_output_format(output_format)

        def filename_mapper(_output, _key):
            return {
                ExportFormat.NATURAL: self._get_file_path('result.json'),
                ExportFormat.FLAT: self._get_file_path(_key + '.json'),
                ExportFormat.PARQUET: self._get_file_path(_key + '.pq'),
                ExportFormat.CSV: self._get_file_path(_key + '.csv'),
                ExportFormat.ARROW: self._get_file_path(_key + '.arrow')
            }[_output]

        format_writers = {
            ExportFormat.NATURAL: self._write_result_as_file,
            ExportFormat.FLAT: self._write_result_as_file,
            ExportFormat.PARQUET: lambda _filename, data: data.to_parquet(_filename),
            ExportFormat.CSV: lambda _filename, data: data.to_csv(_filename, index=False),
            ExportFormat.ARROW: lambda _filename, data: pa.ipc.new_file(_filename, data.schema).write(
                data)
        }
        filenames = []
        if self.result_format == ExportFormat.NATURAL:
            filename = filename_mapper(self.result_format, '')
            format_writers[self.result_format](filename, self.native_result)
            return filename
        for key in self.result:
            filename = filename_mapper(self.result_format, key)
            filenames.append(filename)
            format_writers[self.result_format](filename, self.result[key])
        return filenames

    def track_pg_tables(self, tables, source="default", schema=None):
        """
        Args:
            tables: A list of table names to be tracked in Hasura.
            source: The name of the data source to be used for tracking the tables. Default value is "default".
            schema: The schema name of the tables. If provided, it will be used to qualify the table names.

        Returns:
            A dictionary containing the result of tracking each table. The keys are the table names and the values are the parsed JSON responses.

        """
        result = {}
        for table_name in tables:
            data = {
                "type": "pg_track_table",
                "args": {
                    "source": source,
                    "name": table_name,
                    "configuration": {
                        "apollo_federation_config": {
                            "enable": "v1"
                        }
                    }
                }
            }
            if schema is not None:
                data["args"]["schema"] = schema
            response = requests.post(
                self.uri.replace('graphql', 'metadata'),
                json=data,
                headers={
                    'x-hasura-admin-secret': self.admin_secret
                })
            result[table_name] = json.loads(response.text)
        requests.get(self.uri.replace('graphql', 'reload-schema'))
        self.get_metadata()
        return result

    def replace_metadata(self, metadata=None, allow_inconsistent_metadata=True, allow_warnings=True):
        """
        Args:
            metadata: A dictionary containing the new metadata to be replaced. If None, the existing metadata in the class instance will be used.
            allow_inconsistent_metadata: A boolean indicating whether to allow inconsistent metadata. Defaults to True.
            allow_warnings: A boolean indicating whether to allow warnings. Defaults to True.

        Returns:
            The response in JSON format received from the server after the metadata replacement.

        """
        data = {
            "type": "replace_metadata",
            "version": 2,
            "args": {
                "allow_inconsistent_metadata": allow_inconsistent_metadata,
                "allow_warnings": allow_warnings,
                "metadata": metadata or self.metadata
            }
        }

        response = requests.post(
            self.uri.replace('graphql', 'metadata'),
            json=data,
            headers={
                'x-hasura-admin-secret': self.admin_secret
            })
        result = response.json()
        self.get_metadata()
        return result

    def add_dbml_model_as_source(self, dbml_model_file, configuration, kind, output_file=None, customization=None):
        """
        Args:
            dbml_model_file: A string representing the file path of the DBML model file.
            configuration: A dictionary representing the Hasura data source configuration.
            kind: A string representing the kind of source to be added.
            output_file: (Optional) A string representing the file path where the result will be written. If not provided, the result will not be written to a file.
            customization: (Optional) A dictionary representing the customization options for the source.

        Returns:
            result: A dictionary representing the result of adding the DBML model as a source.

        Example usage:
            dbml_model_file = "/path/to/dbml_model_file.dbml"
            configuration = {
                "connection_info": {
                    "database_url": {
                        "from_env": "PG_DATABASE_URL1"
                    },
                    "isolation_level": "read-committed",
                    "use_prepared_statements": False
                }
            }
            kind = "database"
            output_file = "/path/to/output_file.json"
            customization = {
                "root_fields": {
                    "namespace": "myapp_",
                    "prefix": "prefix_",
                    "suffix": "_suffix"
                },
                "type_names": {
                    "prefix": "Type_",
                    "suffix": "_Type"
                }
            }

            result = add_dbml_model_as_source(dbml_model_file, configuration, kind, output_file, customization)
        """
        result = add_dbml(
            dbml_model_file,
            configuration=configuration,
            kind=kind,
            logging=self.logging,
            hasura_client=self,
            customization=customization
        )
        if output_file is not None:
            self._write_result_as_file(output_file, result)
        return result

    def get_schema_relationships(self):
        rel = Relationships(
            self.get_metadata(),
            self.get_schema(),
            logging_=self.logging,
            client=self
        )
        return rel.get_schema_relationships()

    def metadata_to_neo4j(self, uri, username, password, nodes=None, relationships=None):
        rel = Relationships(
            self.get_metadata(),
            self.get_schema(),
            logging_=self.logging,
            client=self
        )
        if nodes is None or relationships is None:
            nodes, relationships = rel.get_schema_relationships()
        rel.metadata_to_neo4j(uri, username, password, nodes, relationships)



import re
from enum import Enum

from nltk import word_tokenize
from nltk.corpus import wordnet, wordnet as wn
import nltk

nltk.download('wordnet')
nltk.download('punkt')
nltk.download('averaged_perceptron_tagger')


class PartOfSpeech(Enum):
    Noun = 'n'
    Verb = 'v'
    Adjective = 'a|s'
    Conjunction = 'c'


class SpeechTags(Enum):
    Noun = 'NN'
    NounPlural = 'NNS'
    ProperNoun = 'NNP'
    ProperNounPlural = 'NNPS'
    Verb = 'VB'
    VerbThirdPerson = 'VBZ'
    Adjective = 'JJ'
    Adverb = 'RB'
    Preposition = 'IN'
    Determiner = 'DT'
    PersonalPronoun = 'PRP'
    Conjunction = 'CC'
    PossessiveEnding = 'POS'


def get_synonyms(word, part_of_speech=PartOfSpeech.Noun):
    """
    Args:
        word: The word for which you want to find synonyms.
        part_of_speech: The part of speech of the word. Default value is PartOfSpeech.Noun.

    Returns:
        A list of synonyms for the given word and part of speech.

    """
    synonyms = set()
    synsets = wordnet.synsets(word)
    pattern = re.compile(f'^({part_of_speech.value})$') if part_of_speech is not None else re.compile(f'')
    for synset in [syn for syn in synsets if part_of_speech is None or pattern.match(syn.pos())]:
        for lemma in synset.lemmas():
            synonyms.add(lemma.name())
    return list(synonyms)


def get_last_part_of_speech(words=None, part_of_speech=PartOfSpeech.Noun):
    if words is None:
        return None
    for word in reversed(words):
        if is_part_of_speech(word, part_of_speech):
            return word
    return None


def get_last_noun_sequence(words=None):
    if words is None:
        return None
    tokens = word_tokenize(' '.join(words))
    tagged_words = nltk.pos_tag(tokens)
    noun_tags = [SpeechTags.Noun.value, SpeechTags.NounPlural.value, SpeechTags.ProperNoun.value,
                 SpeechTags.ProperNounPlural.value]
    noun_list = []
    for word, tag in reversed(tagged_words):
        if len(noun_list) == 0 and tag in noun_tags:
            noun_list.append(word)
        elif len(noun_list) > 0 and tag not in noun_tags:
            return list(reversed(noun_list))
        elif len(noun_list) > 0 and tag in noun_tags:
            noun_list.append(word)
    return list(reversed(noun_list))


def split_cased_phrase(phrase):
    """
    Splits a camel-cased, snake-cased, or kebab-cased phrase into individual words.

    Args:
        phrase (str): The input phrase.

    Returns:
        list: A list of individual words.
    """
    # Split by underscores (snake case) or hyphens (kebab case)
    words = re.split(r'[_-]', phrase)

    # If the phrase is camel case, further split by capital letters
    if len(words) == 1:
        # Use a regex to find capital letters and split accordingly
        matches = re.finditer(r'(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])', phrase)
        split_string = []
        previous = 0
        for match in matches:
            split_string.append(phrase[previous:match.start()])
            previous = match.start()
        split_string.append(phrase[previous:])  # Add the last word
        words = split_string

    return words


def is_part_of_speech(word, part_of_speech=PartOfSpeech.Noun):
    pattern = re.compile(f'^({part_of_speech.value})$')
    for synset in wn.synsets(word):
        if pattern.match(synset.pos()):
            return True
    return False



import glob
import os
from enum import Enum
from pathlib import Path

import chardet
import numpy as np
import pandas as pd
import camelcaser as cc

from price_parser import Price
from sqlalchemy import create_engine


class Casing(Enum):
    none = None
    snake = 'Snake'
    camel = 'Camel'
    pascal = 'Pascal'

def predict_encoding(file_path: Path, n_lines: int = 20) -> str:
    with file_path.open('rb') as f:
        rawdata = b''.join([f.readline() for _ in range(n_lines)])
    result = chardet.detect(rawdata)
    return result['encoding']


def is_currency_string(input_string):
    try:
        price = Price.fromstring(str(input_string))
        return price.amount is not None and price.currency is not None
    except ValueError:
        return False


def get_currency_amount(input_string):
    try:
        if not pd.isnull(input_string):
            price = Price.fromstring(input_string)
            return price.amount
        else:
            return np.nan
    except ValueError:
        return np.nan


def fix_dates(df: pd.DataFrame, threshold=.1) -> pd.DataFrame:
    # Identify date columns and convert to timestamp
    for col in df.columns:
        if df[col].dtype == 'object':  # Check if the column contains string values
            valid_dates_mask = pd.to_datetime(df[col], errors='coerce').notnull()

            # Calculate the percentage of valid dates
            valid_dates_percentage = valid_dates_mask.sum() / len(df)

            if valid_dates_percentage > threshold:
                # Convert valid dates to datetime format
                df[col] = pd.to_datetime(df[col])
    return df


def fix_currency(df: pd.DataFrame) -> pd.DataFrame:
    for col in df.columns[1:]:
        if df[col].dtype == 'object' and df[col].apply(is_currency_string).all():
            # Remove currency symbols and commas
            df[col] = df[col].apply(get_currency_amount).fillna(df[col])
            df[col] = df[col].astype(float)
    return df


def import_tables(csv_folder: str, uri: str, if_exists='replace', logging=None, casing=Casing.none) -> list[str]:
    """
    Imports tables from CSV files into a database using SQLAlchemy.

    Args:
        csv_folder (str): The path to the folder containing the CSV files.
        uri (str): The URI of the database.
        if_exists (str, optional): The action to take if the table already exists in the database. \
        Defaults to 'replace'. 'append' can also be used.
        logging (Logger, optional): The logging object to use for logging messages. Defaults to None.
        casing (Casing, optional): The casing style to apply to column names. Defaults to Casing.none.

    Returns:
        pd.DataFrame: A DataFrame containing the table names that were imported.

    """
    engine = create_engine(uri)
    logging.info("Connected to database")
    results = []

    # Loop through each CSV file
    for filename in glob.glob(os.path.join(csv_folder, "*.csv")):
        table_name = cc.make_camel_case(os.path.splitext(os.path.basename(filename))[0])
        results.append(table_name)
        df = pd.read_csv(filename, encoding=predict_encoding(Path(filename)))
        df = fix_dates(df)
        df = fix_currency(df)
        if casing == Casing.camel:
            df.columns = [cc.make_lower_camel_case(col) for col in df.columns]
        elif casing == Casing.snake:
            df.columns = [cc.make_snake_case(col.replace(' ', '_')) for col in df.columns]
        elif casing == Casing.pascal:
            df.columns = [cc.make_camel_case(col.replace(' ', '_')) for col in df.columns]
        df.to_sql(table_name, engine, if_exists=if_exists, index=False)
        logging.info(f"Data loaded into '{table_name}'")

    logging.info("All tables created and data loaded successfully!")
    return results



def get_ordinal_of_smallest_number(numbers):
    m = None
    n = None
    for i, x in enumerate(numbers):
        if m is None:
            m = x
            n = i
        if x != 0 and x < m:
            m = x
            n = i
    return n


def compute_deltas(numbers):
    """
    Computes the deltas (differences) between consecutive elements in an array of numbers.

    Args:
        numbers (list or numpy.ndarray): Input array of numbers.

    Returns:
        list: Array of deltas (differences) between consecutive elements.
    """
    deltas = []
    if len(numbers) <= 1:
        deltas.append(0)
    else:
        for i in range(1, len(numbers)):
            delta = numbers[i] - numbers[i - 1]
            deltas.append(abs(delta))
    return deltas
